/*--------------------------------------------------------------------------+
|                                                                          |
| Copyright 2022 fortiss GmbH						                       |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.ext.variability.ui.editor;

import static org.eclipse.swt.SWT.LEFT;
import static org.eclipse.swt.SWT.NONE;
import static org.eclipse.swt.SWT.TOP;
import static org.fortiss.tooling.kernel.ui.util.DataBindingUtils.DECORATION_KEY;
import static org.fortiss.tooling.kernel.ui.util.DataBindingUtils.performCellTextBinding;
import static org.fortiss.tooling.kernel.ui.util.ObservableUtils.observeValue;

import org.eclipse.core.databinding.Binding;
import org.eclipse.core.databinding.DataBindingContext;
import org.eclipse.core.databinding.observable.value.IObservableValue;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.jface.fieldassist.ControlDecoration;
import org.eclipse.jface.viewers.CellEditor;
import org.eclipse.jface.viewers.ColumnViewer;
import org.eclipse.jface.viewers.EditingSupport;
import org.eclipse.jface.viewers.TextCellEditor;
import org.eclipse.jface.viewers.ViewerCell;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Text;
import org.fortiss.tooling.base.ui.fieldassist.FieldAssist;
import org.fortiss.tooling.ext.variability.ui.util.PresenceConditionProposalProvider;
import org.fortiss.tooling.ext.variability.ui.util.PresenceConditionToStringConverter;
import org.fortiss.tooling.ext.variability.ui.util.StringToPresenceConditionConverter;
import org.fortiss.tooling.kernel.ui.databinding.AbstractTextCellDatabindingEditingSupport;
import org.fortiss.variability.model.presence.PresenceCondition;
import org.fortiss.variability.model.presence.PresencePackage;

/**
 * Base class for EditingSupport for {@link PresenceCondition}s.
 * 
 * @author bayha
 */
public abstract class PresenceConditionEditingSupportBase
		extends AbstractTextCellDatabindingEditingSupport {

	/**
	 * A context able to be used for compiling {@link PresenceCondition}s. (Any
	 * EObject in the FileProject.)
	 */
	EObject pcCompilerContext;

	/** Constructor. */
	public PresenceConditionEditingSupportBase(ColumnViewer viewer, DataBindingContext dbc,
			EObject context) {
		super(viewer, dbc);
		this.pcCompilerContext = context;
	}

	/** {@inheritDoc} */
	@Override
	protected CellEditor createCellEditor(final ColumnViewer viewer) {
		Composite control = (Composite)getViewer().getControl();
		final TextCellEditor textCellEditor = new TextCellEditor(control, NONE);
		final Text text = (Text)textCellEditor.getControl();
		text.setData(DECORATION_KEY, new ControlDecoration(text, LEFT | TOP));
		createTextFieldAssist(text);

		text.moveAbove(control);
		return textCellEditor;
	}

	/**
	 * Creates a {@link FieldAssist} for the given {@link Text}. This method is used
	 * for the {@link Text} in this {@link EditingSupport}.
	 * 
	 * @param text
	 *            The {@link Text} to add a {@link FieldAssist} to.
	 */
	protected void createTextFieldAssist(Text text) {
		new FieldAssist(text, new PresenceConditionProposalProvider(pcCompilerContext));
	}

	/** {@inheritDoc} */
	@Override
	protected abstract boolean canEdit(Object element);

	/** {@inheritDoc} */
	@Override
	protected void saveCellEditorValue(final CellEditor cellEditor, final ViewerCell cell) {
		super.saveCellEditorValue(cellEditor, cell);
		getViewer().update(cell.getElement(), null);
	}

	/** {@inheritDoc} */
	@Override
	protected Binding[] createBinding(CellEditor cellEditor, ViewerCell cell, CellEditor editor,
			DataBindingContext context) {
		IObservableValue<?> observerable = getIObservableValue(cell);

		return performCellTextBinding(context, cellEditor.getControl(), observerable,
				new PresenceConditionToStringConverter(),
				new StringToPresenceConditionConverter(pcCompilerContext), null);
	}

	/**
	 * Retrieves {@link IObservableValue} for this editing support.
	 * 
	 * @param cell
	 *            {@link ViewerCell} for creating {@link IObservableValue}.
	 * @return Newly created {@link IObservableValue}.
	 */
	protected IObservableValue<?> getIObservableValue(ViewerCell cell) {
		return observeValue((EObject)cell.getElement(),
				PresencePackage.Literals.HAS_PRESENCE_CONDITION__PRESENCE_CONDITION);
	}

	/** {@inheritDoc} */
	@Override
	protected void setupCellEditor(CellEditor cellEditor) {
		// EMPTY
	}
}
